"""
a dask array class (unyt_dask_array) and helper functions for unyt.



"""

import sys
from functools import wraps

import numpy as np

import unyt.array as ua

if "pytest" in sys.modules:
    # should only happen if pytest is installed *and* already imported,
    # so we can skip collecting doctests from this module when dask isn't installed
    # while avoiding making pytest itself a hard dependency to this module.
    # This check is constructed to work with direct invocation (pytest unyt)
    # as well as through python -m pytest
    import pytest

    pytest.importorskip("dask")
    del pytest

from dask.array.core import Array as DaskArray, finalize as dask_finalize  # noqa: E402

# the following attributes hang off of dask.array.core.Array and do not modify units
_use_unary_decorator = {
    "min",
    "max",
    "sum",
    "mean",
    "std",
    "cumsum",
    "squeeze",
    "rechunk",
    "clip",
    "view",
    "swapaxes",
    "round",
    "copy",
    "repeat",
    "astype",
    "reshape",
    "topk",
}


def _unary_dask_decorator(dask_func, current_unyt_dask):
    """
    a decorator for the simpler dask functions that can just copy over the current
    unit info after applying the dask function. this includes functions that return
    single values (e.g., min()) or functions that change the array in ways that do
    not affect units like reshaping or rounding.

    Parameters
    ----------

    dask_func: func handle
       the dask function handle to call
    current_unyt_dask: unyt_dask_array
       the current instance of a unyt_dask_array

    Returns a new unyt_dask_array instance with appropriate units
    """

    @wraps(dask_func)
    def wrapper(*args, **kwargs):
        da = dask_func(*args, **kwargs)  # will return standard dask array
        return _create_with_quantity(da, current_unyt_dask._unyt_array)

    return wrapper


# the following list are the unyt_quantity/array attributes that will also
# be attributes of the unyt_dask_array. These methods are wrapped with a unit
# conversion decorator.
_unyt_funcs_to_track = {"to", "in_units", "in_cgs", "in_base", "in_mks"}


# helper sanitizing functions for handling ufunc, array operations
def _extract_unyt(obj):
    # returns the hidden _unyt_quantity if it exists in obj, otherwise return obj
    if _is_iterable(obj):
        return [_extract_unyt(ob) for ob in obj]
    return obj._unyt_array if isinstance(obj, unyt_dask_array) else obj


def _extract_dask(obj):
    # returns a plain dask array if the obj is a unyt_dask_array, otherwise return obj
    if _is_iterable(obj):
        return [_extract_dask(ob) for ob in obj]
    return obj.to_dask() if isinstance(obj, unyt_dask_array) else obj


def _extract_unyt_val(obj):
    # returns the value of a unyt_quantity if obj is a unyt_quantity
    if _is_iterable(obj):
        return [_extract_unyt_val(ob) for ob in obj]
    return obj.to_value() if isinstance(obj, ua.unyt_quantity) else obj


def _check_unyt_inputs(ui_0, ui_1):
    # returns True if the args need to be converted to the same unit, based on
    # if the units have the same dimensions but different units.

    if hasattr(ui_0, "units") and hasattr(ui_1, "units"):
        un_0, un_1 = ui_0.units, ui_1.units
        return un_0 != un_1 and un_0.dimensions == un_1.dimensions
    return False


def _is_iterable(obj):
    return type(obj) is tuple or type(obj) is list  # noqa E721


def _sanitize_unit_args(*input):
    # returns sanitized inputs and unyt_inputs for calling the ufunc
    unyt_inputs = _extract_unyt(input)

    if len(unyt_inputs) > 1:
        # note 1: even though we rely on the ufunc applied to the unyt quantities
        # to get the final units of our unyt_dask_array after an operation,
        # we need to first ensure that if our arguments have the same dimensions
        # they are in the same units. This happens internally for unyt_quantities
        # but we also need to apply those internal unit conversions to our
        # dask_unyt_array objects, so we do those checks manually here.
        # note 2: we do NOT check for validity of the operation here. The subsequent
        # call to the ufunc with the unyt_inputs will enforce the unyt rules
        # (e.g., addition must have same dimensions).
        ui_0, ui_1 = unyt_inputs[0], unyt_inputs[1]
        if _check_unyt_inputs(ui_0, ui_1):
            # convert to the unit with the larger base
            input = list(input)
            if ui_0.units.base_value < ui_1.units.base_value:
                input[0] = input[0].to(ui_1.units)
            else:
                input[1] = input[1].to(ui_0.units)
            unyt_inputs = _extract_unyt(input)

    return input, unyt_inputs


def _prep_ufunc(ufunc, *input, extract_dask=False, **kwargs):
    # this function:
    # (1) sanitizes inputs for calls to __array_func__, __array__ and _elementwise
    # (2) applies the function to the hidden unyt quantities
    # (3) (optional) makes inputs extra clean: converts unyt_dask_array args to
    #     plain dask array objects

    # apply the operation to the hidden unyt_quantities
    input, unyt_inputs = _sanitize_unit_args(*input)
    unyt_result = ufunc(*unyt_inputs, **kwargs)

    if extract_dask:
        input = _extract_dask(input)

    input = _extract_unyt_val(input)
    return input, unyt_result


def _post_ufunc(dask_superfunk, unyt_result):
    # a decorator to attach hidden unyt quantity to result of a ufunc, array or
    # elemwise calculation
    def wrapper(*args, **kwargs):
        dask_result = dask_superfunk(*args, **kwargs)
        if hasattr(unyt_result, "units"):
            return _create_with_quantity(dask_result, unyt_result)
        return dask_result

    return wrapper


def _special_dec(the_func):
    # decorator for special operations like __mul__ , __truediv__, which will
    # bypass __getattribute__
    def wrapper(self, *args, **kwargs):
        funcname = the_func.__name__
        ufunc = getattr(ua.unyt_quantity, funcname)
        ufunc_args = (self,) + args
        args, unyt_result = _prep_ufunc(ufunc, *ufunc_args, extract_dask=True, **kwargs)

        # remove the first arg, cause we need to supply self as first arg
        _ = args.pop(0)
        daskresult = the_func(self, *args, **kwargs)

        if hasattr(unyt_result, "units"):
            return _create_with_quantity(daskresult, unyt_result)
        return daskresult  # __lt__, __le__, etc. will hit this

    return wrapper


class unyt_dask_array(DaskArray):
    """
    a dask.array.core.Array subclass that tracks units. This class is only
    recommended for advanced usage, most cases should use the unyt_from_dask
    helper function to instantiate a new unyt_dask_array.

    Parameters
    ----------

    The following parameters are those for a standard dask.array.core.Array:

    dask : dict
        Task dependency graph
    name : string
        Name of array in dask
    shape : tuple of ints
        Shape of the entire array
    chunks: iterable of tuples
        block sizes along each dimension
    dtype : str or dtype
        Typecode or data-type for the new Dask Array
    meta : empty ndarray
        empty ndarray created with same NumPy backend, ndim and dtype as the
        Dask Array being created (overrides dtype)

    The following keyword-only parameters are the same as for a standard
    unyt.unyt_array (check there for definitions):

    units
    registry
    bypass_validation
    unyt_name

    """

    def __new__(
        cls,
        dask_graph,
        name,
        chunks,
        dtype=None,
        meta=None,
        shape=None,
        *,
        units=None,
        registry=None,
        bypass_validation=False,
        unyt_name=None,
    ):
        # get the base dask array
        obj = super().__new__(
            cls,
            dask_graph,
            name,
            chunks,
            dtype,
            meta,
            shape,
        )

        # attach our unyt sidecar quantity
        dtype = obj.dtype
        obj._unyt_array = ua.unyt_array(
            [1.0],
            units,
            registry,
            dtype,
            bypass_validation=bypass_validation,
            name=unyt_name,
        )

        obj.units = obj._unyt_array.units
        obj.unyt_name = obj._unyt_array.name

        # set the unit conversion attributes so they are discoverable. no name
        # conflicts for now, but this could be an issue if _unyt_funcs_to_track
        # is expanded.
        for attr in _unyt_funcs_to_track:
            setattr(obj, attr, getattr(obj._unyt_array, attr))

        return obj

    def _elemwise(self, ufunc, *args, **kwargs):
        args, unyt_result = _prep_ufunc(ufunc, *args, **kwargs)
        return _post_ufunc(super()._elemwise, unyt_result)(ufunc, *args, **kwargs)

    def __array_ufunc__(self, numpy_ufunc, method, *inputs, **kwargs):
        inputs, unyt_result = _prep_ufunc(
            numpy_ufunc, *inputs, extract_dask=True, **kwargs
        )
        wrapped_func = _post_ufunc(super().__array_ufunc__, unyt_result)
        return wrapped_func(numpy_ufunc, method, *inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        args, unyt_result = _prep_ufunc(func, *args, extract_dask=True, **kwargs)
        types = [type(i) for i in args]
        wrapped_func = _post_ufunc(super().__array_function__, unyt_result)
        return wrapped_func(func, types, args, kwargs)

    def __getitem__(self, index):
        return _unary_dask_decorator(super().__getitem__, self)(index)

    def __setitem__(self, key, value):
        # requires dask >= 2021.4.1
        super().__setitem__(key, value)

    def __repr__(self):
        disp_str = super().__repr__().replace("dask.array", "unyt_dask_array")
        units_str = f", units={self.units}>"
        return disp_str.replace(">", units_str)

    def _repr_html_(self):
        # controls jupyter notebook display of an array. called from _repr_html_
        base_table = super()._repr_html_()
        table = base_table.split("\n")
        new_table = []
        for row in table:
            if "</tbody>" in row:
                u = self.units
                newrow = f"    <tr><th> Units </th><td> {u} </td> <td> {u} </td></tr>"
                new_table.append(newrow)
            new_table.append(row)
        return "\n".join(new_table)

    def to_dask(self):
        """
        convert to a plain dask array

        Returns
        -------
        dask.array object

        Examples
        --------

        >>> from unyt import dask_array
        >>> import dask.array as da
        >>> x = da.random.random((10000, 10000), chunks=(1000, 1000))
        >>> x_da = dask_array.unyt_from_dask(x, 'm')
        >>> x_da.to_dask()
        ... # doctest: +NORMALIZE_WHITESPACE
        dask.array<random_sample, shape=(10000, 10000), dtype=float64,
             chunksize=(1000, 1000), chunktype=numpy.ndarray>
        """
        (_, args) = super().__reduce__()
        return DaskArray(*args)

    def __reduce__(self):
        (_, args) = super().__reduce__()
        unyt_state = {
            "_unyt_array": self._unyt_array,
            "units": self.units,
            "unyt_name": self.unyt_name,
        }
        # reminder: the 3rd object returned by a __reduce__ tuple is the object
        # state. when there is no __setstate__, it must be a dict and it will be
        # added to the __dict__ attribute.
        return unyt_dask_array, args, unyt_state

    def __getattribute__(self, name):
        if name in _unyt_funcs_to_track:
            return self._track_conversion(name)

        result = super().__getattribute__(name)
        if name in _use_unary_decorator:
            return _unary_dask_decorator(result, self)

        return result

    def _track_conversion(self, unyt_func_name):
        """
        a decorator to use with unyt functions that convert units

        Parameters
        ----------

        unyt_func_name: str
            the name of the function to call from the sidecar _unyt_quantity. Must be
            an attribute of unyt.unyt_quantity.
        current_unyt_dask: unyt_dask_array
            the current instance of a unyt_dask_array


        Returns a new unyt_dask_array instance with appropriate units
        """

        def wrapper(*args, **kwargs):
            # current value of sidecar quantity
            init_val = self._unyt_array.value

            # get the unyt function handle and call it
            the_func = getattr(self._unyt_array, unyt_func_name)
            new_unyt_quantity = the_func(*args, **kwargs)

            # calculate the conversion factor and pull out the new name and units
            # might be able to use _get_conversion_factor here too...
            factor = new_unyt_quantity.value / init_val

            # apply the factor, return new
            new_unyt_quantity[:] = 1  # reset sidecar value
            new_obj = _create_with_quantity(self * factor, new_unyt_quantity)

            return new_obj

        # functools wrap fails here because unyt_func_name is a string, copy manually:
        wrapper.__doc__ = getattr(self._unyt_array, unyt_func_name).__doc__
        return wrapper

    def __dask_postcompute__(self):
        # a dask hook to catch after .compute(), see
        # https://docs.dask.org/en/latest/custom-collections.html#example-dask-collection
        return _finalize_unyt, ((self.units,))

    def _set_unit_state(self, units, new_unyt_array, unyt_name):
        # sets just the unit state of the object
        self.units = units
        self._unyt_array = new_unyt_array
        self.unyt_name = unyt_name

    # These methods bypass __getattribute__ and numpy hooks, so they are defined
    # explicitly here (but are handled generically by the _special_dec decorator).

    @_special_dec
    def __abs__(self):
        return super().__abs__()

    @_special_dec
    def __pow__(self, other):
        return super().__pow__(other)

    @_special_dec
    def __mul__(self, other):
        return super().__mul__(other)

    @_special_dec
    def __rmul__(self, other):
        return super().__rmul__(other)

    @_special_dec
    def __truediv__(self, other):
        return super().__truediv__(other)

    @_special_dec
    def __rtruediv__(self, other):
        return super().__rtruediv__(other)

    @_special_dec
    def __add__(self, other):
        return super().__add__(other)

    @_special_dec
    def __sub__(self, other):
        return super().__sub__(other)

    @_special_dec
    def __lt__(self, other):
        return super().__lt__(other)

    @_special_dec
    def __le__(self, other):
        return super().__le__(other)

    @_special_dec
    def __gt__(self, other):
        return super().__gt__(other)

    @_special_dec
    def __ge__(self, other):
        return super().__ge__(other)

    @_special_dec
    def __eq__(self, other):
        return super().__eq__(other)

    @_special_dec
    def __ne__(self, other):
        return super().__ne__(other)

    def prod(self, *args, **kwargs):
        _, unit = ua._apply_power_mapping(
            np.multiply, self._unyt_array.units, self.size, self.shape, kwargs
        )
        dask_result = super().prod(*args, **kwargs)
        return _create_with_quantity(dask_result, ua.unyt_array([1.0], unit))


def _finalize_unyt(results, unit_name):
    """
    the function to call from the __dask_postcompute__ hook.

    Parameters
    ----------
    results : the dask results object
    unit_name : the units of the result

    Returns
    -------
    unyt_array or unyt_quantity

    """

    # here, we first call the standard finalize function for a dask array
    # and then return a standard unyt_array from the now in-memory result if
    # the result is an array, otherwise return a unyt_quantity.
    result = dask_finalize(results)

    if type(result) is np.ndarray:
        return ua.unyt_array(result, unit_name)
    else:
        return ua.unyt_quantity(result, unit_name)


def _create_with_quantity(dask_array, new_unyt_array):
    """
    this function instantiates a new unyt_dask_array instance and then sets
    the unit state, including units. Used to wrap dask operations

    Parameters
    ----------
    dask_array : a standard dask array or a unyt_dask_array
    new_unyt_array : a standard unyt array
    remaining arguments get passed to unyt.unyt_array, check there for a
    description.
    """
    if isinstance(dask_array, unyt_dask_array):
        dask_array = dask_array.to_dask()
    out = unyt_from_dask(dask_array)

    # attach the unyt_array
    units = new_unyt_array.units
    unyt_name = new_unyt_array.name

    out._set_unit_state(units, new_unyt_array, unyt_name)
    return out


def unyt_from_dask(
    dask_array,
    units=None,
    *,
    registry=None,
    bypass_validation=False,
    unyt_name=None,
):
    """
    creates a unyt_dask_array from a standard dask array. This function will
    create a new copy of the dask graph associated with an array.

    Parameters
    ----------
    dask_array : a standard dask array
    units : String unit name, unit symbol object, or astropy unit
        The units of the array. Powers must be specified using python
        syntax (cm**3, not cm^3).
    registry : :class:`unyt.unit_registry.UnitRegistry`
        The registry to create units from. If units is already associated
        with a unit registry and this is specified, this will be used instead
        of the registry associated with the unit object.
    bypass_validation : boolean
        If True, all input validation is skipped. Using this option may produce
        corrupted, invalid units or array data, but can lead to significant
        speedups in the input validation logic adds significant overhead. If
        set, units *must* be a valid unit object. Defaults to False.
    unyt_name : string
        The name of the array. Defaults to None. This attribute does not propagate
        through mathematical operations, but is preserved under indexing
        and unit conversions.

    Notes
    -----
        All of the above parameters with the exception of the dask_array are
        the same as for a standard unyt.unyt_array.

    Examples
    --------

    >>> from unyt import dask_array, unyt_quantity
    >>> import dask.array as da
    >>> x = da.ones((1000, 1000), chunks=(100, 100))
    >>> x_da = dask_array.unyt_from_dask(x, 'm')
    >>> type(x_da)
    <class 'unyt.dask_array.unyt_dask_array'>
    >>> x_da.units
    m
    >>> x_da.mean().units
    m
    >>> x_da.mean().compute()
    unyt_quantity(1., 'm')
    >>> x_da.to('cm').mean().compute()
    unyt_quantity(100., 'cm')
    >>> (x_da.to('cm')/unyt_quantity(10, 's')).mean().compute()
    unyt_quantity(10., 'cm/s')

    """

    # reduce the dask array to pull out the arguments required for instantiating
    # a new dask.array.core.Array object and then initialize our new unyt_dask
    # array
    (cls, args) = dask_array.__reduce__()

    da = unyt_dask_array(
        *args,
        units=units,
        registry=registry,
        bypass_validation=bypass_validation,
        unyt_name=unyt_name,
    )

    return da


# note: the unyt_dask_array class has no way of catching daskified reductions.
# operations like dask.array.min() get routed through dask.array.reductions.min()
# and will return plain arrays or float/int values.
#
# When these operations exist as attributes, they can be called and will return
# unyt objects. i.e., :
#
# import dask; import unyt
# x_da = unyt_from_dask(dask.array.ones((10, 10), chunks=(2, 2)), unyt.m)
# dask.array.min(x_da).compute()  #  returns a plain float
# x_da.min().compute()  #  returns a unyt quantity
#
# but when the functions do not exist as attributes, like dask.array.nanmin(),
# it is difficult to handle without manually wrapping all of those reductions
# functions and exposing them here. The following function, reduce_with_units,
# is a compromise: it is a simple helper function for calling
# a general dask.array.reductions method on a unyt_dask array to correctly
# handle units.
#
# check https://github.com/yt-project/unyt/issues/269 for further discussion

_nan_ops = {"nansum", "nanmean", "nanmedian", "nanstd", "nanmax", "nanmin", "nancumsum"}
_passthrough_reductions = {"diagonal", "median"}
_allowed_funcs = _nan_ops.union(_passthrough_reductions, _use_unary_decorator)


def reduce_with_units(dask_func, unyt_dask_in, *args, **kwargs):
    """
    Call a dask.array.reduction function and preserve units.

    Parameters
    ----------
    dask_func : function
        a function handle from dask.array.reduction (e.g., dask.array.min,
        dask.array.var, dask.array.nanstd) to call.
    unyt_dask_in : unyt_dask_array
        the unyt dask array, first argument to dask_func
    args:
        any arguments to the dask_func
    kwargs:
        any keyword arguments to the dask_func

    Returns
    -------
    unyt_dask_array:
        the result of dask_func with units preserved

    Examples
    --------
    >>> import dask.array
    >>> from unyt.dask_array import unyt_from_dask, reduce_with_units
    >>> a = dask.array.ones((10000,), chunks=(100,))
    >>> a = unyt_from_dask(a, 'm')
    >>> b = reduce_with_units(dask.array.median, a, axis=0)
    >>> b.compute()
    unyt_quantity(1., 'm')

    """

    if dask_func.__name__ in _allowed_funcs:
        # e.g., min, max, nanstd. functions that return the same units can
        # be called directly as the dask function will treat unyt_dask_in as
        # a standard dask array and then we copy over the initial units.
        return _unary_dask_decorator(dask_func, unyt_dask_in)(
            unyt_dask_in, *args, **kwargs
        )
    else:
        # the operation may change the units
        npfunc = getattr(np, dask_func.__name__, None)
        if npfunc:
            newargs, unyt_result = _prep_ufunc(
                npfunc, unyt_dask_in, *args, extract_dask=True, **kwargs
            )
            return _post_ufunc(dask_func, unyt_result)(*newargs)
        else:
            raise ValueError("could not deduce np equivalent of dask reduction")



"""
A set of convenient on-demand imports
"""

import sys
from functools import wraps
from importlib.util import find_spec


class NotAModule:
    """
    A class to implement an informative error message that will be outputted if
    someone tries to use an on-demand import without having the requisite
    package installed.
    """

    def __init__(self, pkg_name, exc=None):
        self.pkg_name = pkg_name
        self._original_exception = exc
        error_note = (
            f"Something went wrong while trying to lazy-import {pkg_name}. "
            f"Please make sure that {pkg_name} is properly installed.\n"
            "If the problem persists, please file an issue at "
            "https://github.com/yt-project/unyt/issues/new"
        )
        if exc is None:
            self.error = ImportError(error_note)
        elif sys.version_info >= (3, 11):
            exc.add_note(error_note)
            self.error = exc
        else:
            # mimic Python 3.11 behaviour:
            # preserve error message and traceback
            self.error = type(exc)(f"{exc!s}\n{error_note}").with_traceback(
                exc.__traceback__
            )

    def __getattr__(self, item):
        raise self.error

    def __call__(self, *args, **kwargs):
        raise self.error

    def __repr__(self) -> str:
        if self._original_exception is None:
            return f"NotAModule({self.pkg_name!r})"
        else:
            return f"NotAModule({self.pkg_name!r}, {self._original_exception!r})"


class OnDemand:
    _default_factory: type[NotAModule] = NotAModule

    def __init_subclass__(cls):
        if not cls.__name__.endswith("_imports"):
            raise TypeError(f"class {cls}'s name needs to be suffixed '_imports'")

    def __new__(cls):
        if cls is OnDemand:
            raise TypeError("The OnDemand base class cannot be instantiated.")
        else:
            return object.__new__(cls)

    @property
    def _name(self) -> str:
        _name, _, _suffix = self.__class__.__name__.rpartition("_")
        return _name

    @property
    def __is_available__(self) -> bool:
        return find_spec(self._name) is not None


def safe_import(func):
    @property
    @wraps(func)
    def inner(self):
        try:
            return func(self)
        except ImportError as exc:
            return self._default_factory(self._name, exc)

    return inner


class astropy_imports(OnDemand):
    @safe_import
    def log(self):
        from astropy import log

        if log.exception_logging_enabled():
            log.disable_exception_logging()

        return log

    @safe_import
    def units(self):
        from astropy import units

        self.log  # noqa: B018
        return units

    @safe_import
    def __version__(self):
        from astropy import __version__

        return __version__


_astropy = astropy_imports()


class h5py_imports(OnDemand):
    @safe_import
    def File(self):
        from h5py import File

        return File

    @safe_import
    def __version__(self):
        from h5py import __version__

        return __version__


_h5py = h5py_imports()


class pint_imports(OnDemand):
    @safe_import
    def UnitRegistry(self):
        from pint import UnitRegistry

        return UnitRegistry


_pint = pint_imports()


class matplotlib_imports(OnDemand):
    @safe_import
    def __version__(self):
        from matplotlib import __version__

        return __version__

    @safe_import
    def pyplot(self):
        from matplotlib import pyplot

        return pyplot

    @safe_import
    def units(self):
        from matplotlib import units

        return units

    @safe_import
    def use(self):
        from matplotlib import use

        return use


_matplotlib = matplotlib_imports()


class dask_imports(OnDemand):
    @safe_import
    def array(self):
        from dask import array

        return array

    @safe_import
    def __version__(self):
        from dask import __version__

        return __version__


_dask = dask_imports()



from weakref import WeakKeyDictionary

from matplotlib.units import AxisInfo, ConversionInterface

from unyt.array import unyt_array, unyt_quantity
from unyt.unit_object import Unit


class unyt_arrayConverter(ConversionInterface):
    """Matplotlib interface for unyt_array"""

    _instance = None
    _labelstyle = "()"
    _axisnames = WeakKeyDictionary()

    # ensure that unyt_arrayConverter is a singleton
    def __new__(cls):
        if unyt_arrayConverter._instance is None:
            unyt_arrayConverter._instance = super().__new__(cls)
        return unyt_arrayConverter._instance

    # When matplotlib first encounters a type in its units.registry, it will
    # call default_units to obtain the units. Then it calls axisinfo to
    # customize the axis - in our case, just set the label. Then matplotlib calls
    # convert.

    @staticmethod
    def axisinfo(unit, axis):
        """Set the axis label based on unit

        Parameters
        ----------

        unit : Unit object, string, or tuple
            This parameter comes from unyt_arrayConverter.default_units() or from
            user code such as Axes.plot(), Axis.set_units(), etc. In user code, it
            is possible to convert the plotted units by specifying the new unit as
            a string, such as "ms", or as a tuple, such as ("J", "thermal")
            following the call signature of unyt_array.convert_to_units().
        axis : Axis object

        Returns
        -------

        AxisInfo object with the label formatted as in-line math latex
        """
        if isinstance(unit, tuple):
            unit = unit[0]
        unit_obj = unit if isinstance(unit, Unit) else Unit(unit)
        name = unyt_arrayConverter._axisnames.get(axis, "")
        if unit_obj.is_dimensionless:
            label = name
        else:
            name += " "
            unit_str = unit_obj.latex_representation()
            if unyt_arrayConverter._labelstyle == "[]":
                label = name + "$\\left[" + unit_str + "\\right]$"
            elif unyt_arrayConverter._labelstyle == "/":
                axsym = "$q_{\\rm" + axis.axis_name + "}$"
                name = axsym if name == " " else name
                if "/" in unit_str:
                    label = name + "$\\;/\\;\\left(" + unit_str + "\\right)$"
                else:
                    label = name + "$\\;/\\;" + unit_str + "$"
            else:
                label = name + "$\\left(" + unit_str + "\\right)$"
        return AxisInfo(label=label.strip())

    @staticmethod
    def default_units(x, axis):
        """Return the Unit object of the unyt_array x

        Parameters
        ----------

        x : unyt_array
        axis : Axis object

        Returns
        -------

        Unit object
        """
        # In the case where the first matplotlib command is setting limits,
        # x may be a tuple of length two (with the same units).
        if isinstance(x, tuple):
            name = getattr(x[0], "name", "")
            units = x[0].units
        else:
            name = getattr(x, "name", "")
            units = x.units

        # maintain a mapping between Axis and name since Axis does not point to
        # its underlying data and we want to propagate the name to the axis
        # label in the subsequent call to axisinfo
        unyt_arrayConverter._axisnames[axis] = name if name is not None else ""
        return units

    @staticmethod
    def convert(value, unit, axis):
        """Convert the units of value to unit

        Parameters
        ----------

        value : unyt_array, unyt_quantity, or sequence there of
        unit : Unit, string or tuple
            This parameter comes from unyt_arrayConverter.default_units() or from
            user code such as Axes.plot(), Axis.set_units(), etc. In user code, it
            is possible to convert the plotted units by specifying the new unit as
            a string, such as "ms", or as a tuple, such as ("J", "thermal")
            following the call signature of unyt_array.convert_to_units().
        axis : Axis object

        Returns
        -------

        unyt_array

        Raises
        ------

        UnitConversionError if unit does not have the same dimensions as value or
        if we don't know how to convert value.
        """
        converted_value = value
        if isinstance(unit, str) or isinstance(unit, Unit):
            unit = (unit,)
        if isinstance(value, (unyt_array, unyt_quantity)):
            converted_value = value.to(*unit)
        else:
            value_type = type(value)
            converted_value = []
            for obj in value:
                converted_value.append(obj.to(*unit))
            converted_value = value_type(converted_value)
        return converted_value



"""
Tests for support of subclasses of unyt_array and unyt_quantity.
"""

import warnings
from collections.abc import Callable, Collection, Iterable
from importlib.metadata import version
from numbers import Number as numeric_type

import numpy as np
from packaging.version import Version

import unyt
from unyt import unyt_array, unyt_quantity
from unyt._array_functions import (
    _trapezoid_func,
    allclose as unyt_allclose,
    around as unyt_around,
    array2string as unyt_array2string,
    array_equal as unyt_array_equal,
    array_equiv as unyt_array_equiv,
    array_repr as unyt_array_repr,
    block as unyt_block,
    choose as unyt_choose,
    clip as unyt_clip,
    column_stack as unyt_column_stack,
    concatenate as unyt_concatenate,
    convolve as unyt_convolve,
    copyto as unyt_copyto,
    correlate as unyt_correlate,
    cross as unyt_cross,
    diff as unyt_diff,
    dot as unyt_dot,
    dstack as unyt_dstack,
    ediff1d as unyt_ediff1d,
    einsum as unyt_einsum,
    fft_fft as unyt_fft_fft,
    fft_fft2 as unyt_fft_fft2,
    fft_fftn as unyt_fft_fftn,
    fft_fftshift as unyt_fft_fftshift,
    fft_hfft as unyt_fft_hfft,
    fft_ifft as unyt_fft_ifft,
    fft_ifft2 as unyt_fft_ifft2,
    fft_ifftn as unyt_fft_ifftn,
    fft_ifftshift as unyt_fft_ifftshift,
    fft_ihfft as unyt_fft_ihfft,
    fft_irfft as unyt_fft_irfft,
    fft_irfft2 as unyt_fft_irfft2,
    fft_irfftn as unyt_fft_irfftn,
    fft_rfft as unyt_fft_rfft,
    fft_rfft2 as unyt_fft_rfft2,
    fft_rfftn as unyt_fft_rfftn,
    fill_diagonal as unyt_fill_diagonal,
    geomspace as unyt_geomspace,
    histogram as unyt_histogram,
    histogram2d as unyt_histogram2d,
    histogram_bin_edges as unyt_histogram_bin_edges,
    histogramdd as unyt_histogramdd,
    hstack as unyt_hstack,
    inner as unyt_inner,
    insert as unyt_insert,
    interp as unyt_interp,
    intersect1d as unyt_intersect1d,
    isclose as unyt_isclose,
    isin as unyt_isin,
    kron as unyt_kron,
    linalg_det as unyt_linalg_det,
    linalg_eig as unyt_linalg_eig,
    linalg_eigh as unyt_linalg_eigh,
    linalg_eigvals as unyt_linalg_eigvals,
    linalg_eigvalsh as unyt_linalg_eigvalsh,
    linalg_inv as unyt_linalg_inv,
    linalg_lstsq as unyt_linalg_lstsq,
    linalg_pinv as unyt_linalg_pinv,
    linalg_solve as unyt_linalg_solve,
    linalg_svd as unyt_linalg_svd,
    linalg_tensorinv as unyt_linalg_tensorinv,
    linalg_tensorsolve as unyt_linalg_tensorsolve,
    linspace as unyt_linspace,
    logspace as unyt_logspace,
    nanpercentile as unyt_nanpercentile,
    nanquantile as unyt_nanquantile,
    norm as unyt_linalg_norm,  # not linalg_norm, doesn't follow usual pattern
    outer as unyt_outer,
    pad as unyt_pad,
    percentile as unyt_percentile,
    place as unyt_place,
    prod as unyt_prod,
    ptp as unyt_ptp,
    put as unyt_put,
    put_along_axis as unyt_put_along_axis,
    putmask as unyt_putmask,
    quantile as unyt_quantile,
    savetxt as unyt_savetxt,
    searchsorted as unyt_searchsorted,
    select as unyt_select,
    setdiff1d as unyt_setdiff1d,
    sinc as unyt_sinc,
    sort_complex as unyt_sort_complex,
    stack as unyt_stack,
    take as unyt_take,
    tensordot as unyt_tensordot,
    trace as unyt_trace,
    trapezoid as unyt_trapezoid,
    tril as unyt_tril,
    triu as unyt_triu,
    union1d as unyt_union1d,
    unwrap as unyt_unwrap,
    var as unyt_var,
    vdot as unyt_vdot,
    vstack as unyt_vstack,
    where as unyt_where,
)
from unyt.array import _iterable, multiple_output_operators

NUMPY_VERSION = Version(version("numpy"))
if NUMPY_VERSION >= Version("2.0.0dev0"):
    from unyt._array_functions import linalg_outer as unyt_linalg_outer
if NUMPY_VERSION < Version("2.0.0dev0"):
    from unyt._array_functions import asfarray as unyt_asfarray

_HANDLED_FUNCTIONS = {}


class ExtraAttributeError(Exception):
    """
    Raised when function arguments have incompatible ``extra_attr``.
    """

    def __init__(self, message: str = "Tried to mix different extra_attr!") -> None:
        self.message = message


# first we define helper functions to handle repetitive operations in wrapping unyt &
# numpy functions (we will actually wrap the functions below):


def _copy_extra_attr_if_present(
    from_obj: object, to_obj: object, copy_units=False
) -> object:
    """
    Copy extra_attr across two objects.
    """
    if not (
        isinstance(to_obj, subclass_uarray) and isinstance(from_obj, subclass_uarray)
    ):
        return to_obj
    if copy_units:
        to_obj.units = from_obj.units
    if hasattr(from_obj, "extra_attr"):
        to_obj.extra_attr = from_obj.extra_attr
    return to_obj


def _propagate_extra_attr_to_result(func: Callable) -> Callable:
    """
    Wrapper that copies the extra attribute from first input argument to first
    output.

    Many functions take one input (or have a first input that has a close
    correspondence to the output) and one output. This helper copies the
    ``extra_attr`` attribute from the first input argument to the output. Can
    be used as a decorator on functions (the first argument is then the first
    argument of the function) or methods (the first argument is then ``self``).
    If the output is not a ``subclass_uarray`` it is not promoted (and then no
    attributes are copied).
    """

    def wrapped(obj, *args, **kwargs):
        return _copy_extra_attr_if_present(obj, func(obj, *args, **kwargs))

    return wrapped


def _prepare_array_func_args(*args, _default_cm: bool = True, **kwargs) -> dict:
    """
    For this minimalist class we check that the extra_attr matches and raise if not.
    """
    args_extra_attrs = [
        (hasattr(arg, "extra_attr"), getattr(arg, "extra_attr", None)) for arg in args
    ]
    kwarg_extra_attrs = {
        k: (hasattr(kwarg, "extra_attr"), getattr(kwarg, "extra_attr", None))
        for k, kwarg in kwargs.items()
    }
    extra_attr_values_where_present = [ea[1] for ea in args_extra_attrs if ea[0]] + [
        ea[1] for ea in kwarg_extra_attrs.values() if ea[0]
    ]
    # here we check that all of the extra_attr match (could be True, False or None):
    if not len(set(extra_attr_values_where_present)) <= 1:
        raise ExtraAttributeError
    # we could modify the args and kwargs before returning them here to "prepare" them
    return {
        "args": args,
        "kwargs": kwargs,
        "extra_attr": (
            extra_attr_values_where_present[0]
            if len(extra_attr_values_where_present) > 0
            else None
        ),
    }


def _return_helper(
    res: np.ndarray,
    helper_result: dict,
    out: np.ndarray | None = None,
) -> "subclass_uarray":
    """
    Helper function to attach our extra_attr to return values of wrapped
    functions.
    """
    res = _promote_unyt_to_subclass(res)
    if isinstance(res, subclass_uarray):  # also recognizes subclass_uquantity
        res.extra_attr = helper_result["extra_attr"]
    if isinstance(out, subclass_uarray):  # also recognizes subclass_uquantity
        out.extra_attr = helper_result["extra_attr"]
    return res


def _ensure_result_is_subclass_uarray_or_uquantity(func: Callable) -> Callable:
    """
    Wrapper that converts any ``unyt_array`` or ``unyt_quantity``
    instances in function output to subclass equivalents.

    If the wrapped function returns a ``tuple`` (as many numpy functions do) it
    is iterated over (but not recursively) and each element with a unyt class
    type is upgraded to its subclass equivalent. If anything but a ``tuple``
    is returned, that object is promoted to the subclass equivalent if it is
    of a unyt class type.
    """

    def wrapped(*args, **kwargs) -> object:
        # omit docstring so that sphinx picks up docstring of wrapped function
        result = func(*args, **kwargs)
        if isinstance(result, tuple):
            return tuple(
                _ensure_array_or_quantity_matches_shape(_promote_unyt_to_subclass(item))
                for item in result
            )
        else:
            return _ensure_array_or_quantity_matches_shape(
                _promote_unyt_to_subclass(result)
            )

    return wrapped


def _promote_unyt_to_subclass(input_object: object) -> object:
    """
    Upgrades the input unyt instance to its subclass equivalent.
    """
    if isinstance(input_object, unyt_quantity) and not isinstance(
        input_object, subclass_uquantity
    ):
        return input_object.view(subclass_uquantity)
    elif isinstance(input_object, unyt_array) and not isinstance(
        input_object, subclass_uarray
    ):
        return input_object.view(subclass_uarray)
    else:
        return input_object


def _ensure_array_or_quantity_matches_shape(input_object: object) -> object:
    """
    Convert scalars to ``subclass_uquantity`` and arrays to
    ``subclass_uarray``.
    """
    if (
        isinstance(input_object, subclass_uarray)
        and not isinstance(input_object, subclass_uquantity)
        and input_object.shape == ()
    ):
        return input_object.view(subclass_uquantity)
    elif isinstance(input_object, subclass_uquantity) and input_object.shape != ():
        return input_object.view(subclass_uarray)
    else:
        return input_object


def _default_unary_wrapper(unyt_func: Callable) -> Callable:
    """
    Wrapper helper for unary functions with typical behaviour.

    Can be used as a decorator.
    """

    def wrapper(*args, **kwargs):
        helper_result = _prepare_array_func_args(*args, **kwargs)
        res = unyt_func(*helper_result["args"], **helper_result["kwargs"])
        if "out" in kwargs:
            return _return_helper(res, helper_result, out=kwargs["out"])
        else:
            return _return_helper(res, helper_result)

    return wrapper


def _default_binary_wrapper(unyt_func: Callable) -> Callable:
    """
    Wrapper helper for binary functions with typical behaviour.

    Can be used as a decorator.
    """

    def wrapper(*args, **kwargs):
        helper_result = _prepare_array_func_args(*args, **kwargs)
        res = unyt_func(*helper_result["args"], **helper_result["kwargs"])
        if "out" in kwargs:
            return _return_helper(res, helper_result, out=kwargs["out"])
        else:
            return _return_helper(res, helper_result)

    return wrapper


# here an alias, but often comparisons have their own particular default logic:
_default_comparison_wrapper = _default_binary_wrapper


def _default_oplist_wrapper(unyt_func: Callable) -> Callable:
    """
    Wrapper helper for functions accepting a list of operands with typical behaviour.

    Can be used as a decorator.
    """

    def wrapper(*args, **kwargs):
        helper_result = _prepare_array_func_args(*args, **kwargs)
        helper_result_oplist = _prepare_array_func_args(*args[0])
        res = unyt_func(
            helper_result_oplist["args"],
            *helper_result["args"][1:],
            **helper_result["kwargs"],
        )
        return _return_helper(res, helper_result_oplist)

    return wrapper


def implements(numpy_function: Callable) -> Callable:
    """
    Register an __array_function__ implementation for subclass_uarray objects.

    Intended for use as a decorator.
    """

    # See NEP 18 https://numpy.org/neps/nep-0018-array-function-protocol.html
    def decorator(func: Callable) -> Callable:
        """
        Actually register the specified function.
        """
        _HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


# Next we wrap functions from unyt and numpy. There's not much point in writing docstrings
# or type hints for all of these.

# Now we wrap functions that unyt handles explicitly (below that will be those not handled
# explicitly):


@implements(np.array2string)
def array2string(a, *args, **kwargs):
    res = unyt_array2string(a, *args, **kwargs)
    return res + f" extra_attr={a.extra_attr}"


implements(np.dot)(_default_binary_wrapper(unyt_dot))
implements(np.vdot)(_default_binary_wrapper(unyt_vdot))
implements(np.inner)(_default_binary_wrapper(unyt_inner))
implements(np.outer)(_default_binary_wrapper(unyt_outer))
implements(np.kron)(_default_binary_wrapper(unyt_kron))
implements(np.histogram_bin_edges)(_default_unary_wrapper(unyt_histogram_bin_edges))
implements(np.linalg.inv)(_default_unary_wrapper(unyt_linalg_inv))
implements(np.linalg.tensorinv)(_default_unary_wrapper(unyt_linalg_tensorinv))
implements(np.linalg.pinv)(_default_unary_wrapper(unyt_linalg_pinv))
implements(np.linalg.svd)(_default_unary_wrapper(unyt_linalg_svd))


def _histogram(a, bins=10, range=None, density=None, weights=None, normed=None):
    helper_result = _prepare_array_func_args(
        a, bins=bins, range=range, density=density, weights=weights, normed=normed
    )
    kwargs = {
        "bins": helper_result["kwargs"]["bins"],
        "range": helper_result["kwargs"]["range"],
        "density": helper_result["kwargs"]["density"],
        "weights": helper_result["kwargs"]["weights"],
    }
    if NUMPY_VERSION < Version("1.24"):
        kwargs["normed"] = helper_result["kwargs"]["normed"]
    counts, bins = unyt_histogram(*helper_result["args"], **kwargs)
    counts = _promote_unyt_to_subclass(counts)
    if isinstance(counts, subclass_uarray):  # also recognizes subclass_uquantity
        counts.extra_attr = helper_result["extra_attr"]
    return counts, _return_helper(bins, helper_result)


if NUMPY_VERSION >= Version("1.24"):

    @implements(np.histogram)
    def histogram(a, bins=10, range=None, density=None, weights=None):
        return _histogram(a, bins=bins, range=range, density=density, weights=weights)

else:

    @implements(np.histogram)
    def histogram(a, bins=10, range=None, normed=None, density=None, weights=None):
        return _histogram(a, bins=bins, range=range, density=density, weights=weights)


def _histogram2d(x, y, *, bins=10, range=None, density=None, weights=None, normed=None):
    if range is not None:
        xrange, yrange = range
    else:
        xrange, yrange = None, None

    try:
        N = len(bins)
    except TypeError:
        N = 1
    if N != 2:
        xbins = ybins = bins
    elif N == 2:
        xbins, ybins = bins
    helper_result_x = _prepare_array_func_args(x, bins=xbins, range=xrange)
    helper_result_y = _prepare_array_func_args(y, bins=ybins, range=yrange)
    if not density:
        helper_result_w = _prepare_array_func_args(weights=weights, normed=normed)
        if (helper_result_x["kwargs"]["range"] is None) and (
            helper_result_y["kwargs"]["range"] is None
        ):
            safe_range = None
        else:
            safe_range = (
                helper_result_x["kwargs"]["range"],
                helper_result_y["kwargs"]["range"],
            )
        kwargs = {
            "bins": (
                helper_result_x["kwargs"]["bins"],
                helper_result_y["kwargs"]["bins"],
            ),
            "range": safe_range,
            "density": density,
            "weights": helper_result_w["kwargs"]["weights"],
        }
        if NUMPY_VERSION < Version("1.24"):
            kwargs["normed"] = helper_result_w["kwargs"]["normed"]
        counts, xbins, ybins = unyt_histogram2d(
            helper_result_x["args"][0],
            helper_result_y["args"][0],
            **kwargs,
        )
        if weights is not None:
            counts = _promote_unyt_to_subclass(counts)
            if isinstance(
                counts, subclass_uarray
            ):  # also recognizes subclass_uquantity
                counts.extra_attr = helper_result_w["extra_attr"]
    else:  # density=True
        # now x, y and weights must be compatible because they will combine
        # we unpack input to the helper to get everything checked for compatibility
        helper_result = _prepare_array_func_args(
            x,
            y,
            xbins=xbins,
            ybins=ybins,
            xrange=xrange,
            yrange=yrange,
            weights=weights,
            normed=normed,
        )
        if (helper_result["kwargs"]["xrange"] is None) and (
            helper_result["kwargs"]["yrange"] is None
        ):
            safe_range = None
        else:
            safe_range = (
                helper_result["kwargs"]["xrange"],
                helper_result["kwargs"]["yrange"],
            )
        kwargs = {
            "bins": (
                helper_result["kwargs"]["xbins"],
                helper_result["kwargs"]["ybins"],
            ),
            "range": safe_range,
            "density": density,
            "weights": helper_result["kwargs"]["weights"],
        }
        if NUMPY_VERSION < Version("1.24"):
            kwargs["normed"] = helper_result["kwargs"]["normed"]
        counts, xbins, ybins = unyt_histogram2d(
            helper_result["args"][0], helper_result["args"][1], **kwargs
        )
        counts = _promote_unyt_to_subclass(counts)
        if isinstance(counts, subclass_uarray):  # also recognizes subclass_uquantity
            counts.extra_attr = helper_result["extra_attr"]
    return (
        counts,
        _return_helper(xbins, helper_result_x),
        _return_helper(ybins, helper_result_y),
    )


if NUMPY_VERSION >= Version("1.24"):

    @implements(np.histogram2d)
    def histogram2d(x, y, bins=10, range=None, density=None, weights=None):
        return _histogram2d(
            x, y, bins=bins, range=range, density=density, weights=weights
        )

else:

    @implements(np.histogram2d)
    def histogram2d(x, y, bins=10, range=None, normed=None, weights=None, density=None):
        return _histogram2d(
            x,
            y,
            bins=bins,
            range=range,
            normed=normed,
            weights=weights,
            density=density,
        )


def _histogramdd(sample, bins=10, range=None, density=None, weights=None, normed=None):
    D = len(sample)
    if range is not None:
        ranges = range
    else:
        ranges = D * [None]

    if np.isscalar(bins):
        bins = D * [bins]
    helper_results = [
        _prepare_array_func_args(s, bins=b, range=r)
        for s, b, r in zip(sample, bins, ranges)
    ]
    if not density:
        helper_result_w = _prepare_array_func_args(weights=weights)
        all_ranges = [
            helper_result["kwargs"]["range"] for helper_result in helper_results
        ]
        if set(all_ranges) == {None}:
            safe_range = None
        else:
            safe_range = [
                helper_result["kwargs"]["range"] for helper_result in helper_results
            ]
        kwargs = {
            "bins": [
                helper_result["kwargs"]["bins"] for helper_result in helper_results
            ],
            "range": safe_range,
            "density": density,
            "weights": helper_result_w["kwargs"]["weights"],
        }
        if NUMPY_VERSION < Version("1.24"):
            kwargs["normed"] = normed
        counts, bins = unyt_histogramdd(
            [helper_result["args"][0] for helper_result in helper_results],
            **kwargs,
        )
        if weights is not None:
            counts = _promote_unyt_to_subclass(counts)
            if isinstance(counts, subclass_uarray):
                counts.extra_attr = helper_result_w["extra_attr"]
    else:  # density=True
        # now sample and weights must be compatible because they will combine
        # we unpack input to the helper to get everything checked for compatibility
        helper_result = _prepare_array_func_args(
            *sample, bins=bins, range=range, weights=weights
        )
        kwargs = {
            "bins": helper_result["kwargs"]["bins"],
            "range": helper_result["kwargs"]["range"],
            "density": density,
            "weights": helper_result["kwargs"]["weights"],
        }
        if NUMPY_VERSION < Version("1.24"):
            kwargs["normed"] = normed
        counts, bins = unyt_histogramdd(
            helper_result["args"],
            **kwargs,
        )
        counts = _promote_unyt_to_subclass(counts)
        if isinstance(counts, subclass_uarray):  # also recognizes subclass_uquantity
            counts.extra_attr = helper_result["extra_attr"]
    return (
        counts,
        tuple(
            _return_helper(b, helper_result)
            for b, helper_result in zip(bins, helper_results)
        ),
    )


if NUMPY_VERSION >= Version("1.24"):

    @implements(np.histogramdd)
    def histogramdd(sample, bins=10, range=None, density=None, weights=None):
        return _histogramdd(
            sample, bins=bins, range=range, density=density, weights=weights
        )

else:

    @implements(np.histogramdd)
    def histogramdd(
        sample, bins=10, range=None, normed=None, weights=None, density=None
    ):
        return _histogramdd(
            sample,
            bins=bins,
            range=range,
            normed=normed,
            weights=weights,
            density=density,
        )


implements(np.concatenate)(_default_oplist_wrapper(unyt_concatenate))
implements(np.cross)(_default_binary_wrapper(unyt_cross))
implements(np.intersect1d)(_default_binary_wrapper(unyt_intersect1d))
implements(np.union1d)(_default_binary_wrapper(unyt_union1d))
implements(np.linalg.norm)(_default_unary_wrapper(unyt_linalg_norm))
implements(np.vstack)(_default_oplist_wrapper(unyt_vstack))
implements(np.hstack)(_default_oplist_wrapper(unyt_hstack))
implements(np.dstack)(_default_oplist_wrapper(unyt_dstack))
implements(np.column_stack)(_default_oplist_wrapper(unyt_column_stack))
implements(np.stack)(_default_oplist_wrapper(unyt_stack))
implements(np.around)(_default_unary_wrapper(unyt_around))


def _prepare_array_block_args(lst, recursing=False):
    """
    Block accepts only a nested list of array "blocks". We need to recurse on this.
    """
    helper_results = []
    if isinstance(lst, list):
        for item in lst:
            if isinstance(item, list):
                helper_results += _prepare_array_block_args(item, recursing=True)
            else:
                helper_results.append(_prepare_array_func_args(item))
    if recursing:
        return helper_results
    eas = [hr["extra_attr"] for hr in helper_results]
    if set(eas) == {True}:
        ret_ea = True
    elif set(eas) == {None}:
        ret_ea = None
    elif set(eas) == {False}:
        ret_ea = False
    else:
        # mixed values
        raise ExtraAttributeError
    ret_lst = lst
    return {
        "args": ret_lst,
        "kwargs": {},
        "extra_attr": ret_ea,
    }


@implements(np.block)
def block(arrays):
    # block is a special case since we need to recurse more than one level
    # down the list of arrays.
    helper_result_block = _prepare_array_block_args(arrays)
    res = unyt_block(helper_result_block["args"])
    return _return_helper(res, helper_result_block)


implements(np.fft.fft)(_default_unary_wrapper(unyt_fft_fft))
implements(np.fft.fft2)(_default_unary_wrapper(unyt_fft_fft2))
implements(np.fft.fftn)(_default_unary_wrapper(unyt_fft_fftn))
implements(np.fft.hfft)(_default_unary_wrapper(unyt_fft_hfft))
implements(np.fft.rfft)(_default_unary_wrapper(unyt_fft_rfft))
implements(np.fft.rfft2)(_default_unary_wrapper(unyt_fft_rfft2))
implements(np.fft.rfftn)(_default_unary_wrapper(unyt_fft_rfftn))
implements(np.fft.ifft)(_default_unary_wrapper(unyt_fft_ifft))
implements(np.fft.ifft2)(_default_unary_wrapper(unyt_fft_ifft2))
implements(np.fft.ifftn)(_default_unary_wrapper(unyt_fft_ifftn))
implements(np.fft.ihfft)(_default_unary_wrapper(unyt_fft_ihfft))
implements(np.fft.irfft)(_default_unary_wrapper(unyt_fft_irfft))
implements(np.fft.irfft2)(_default_unary_wrapper(unyt_fft_irfft2))
implements(np.fft.irfftn)(_default_unary_wrapper(unyt_fft_irfftn))
implements(np.fft.fftshift)(_default_unary_wrapper(unyt_fft_fftshift))
implements(np.fft.ifftshift)(_default_unary_wrapper(unyt_fft_ifftshift))
implements(np.sort_complex)(_default_unary_wrapper(unyt_sort_complex))
implements(np.isclose)(_default_comparison_wrapper(unyt_isclose))
implements(np.allclose)(_default_comparison_wrapper(unyt_allclose))
implements(np.array_equal)(_default_comparison_wrapper(unyt_array_equal))
implements(np.array_equiv)(_default_comparison_wrapper(unyt_array_equiv))
implements(np.linspace)(_default_binary_wrapper(unyt_linspace))
implements(np.logspace)(_default_binary_wrapper(unyt_logspace))
implements(np.geomspace)(_default_binary_wrapper(unyt_geomspace))
implements(np.copyto)(_default_binary_wrapper(unyt_copyto))
implements(np.prod)(_default_unary_wrapper(unyt_prod))
implements(np.var)(_default_unary_wrapper(unyt_var))
implements(np.trace)(_default_unary_wrapper(unyt_trace))
implements(np.percentile)(_default_unary_wrapper(unyt_percentile))
implements(np.quantile)(_default_unary_wrapper(unyt_quantile))
implements(np.nanpercentile)(_default_unary_wrapper(unyt_nanpercentile))
implements(np.nanquantile)(_default_unary_wrapper(unyt_nanquantile))
implements(np.linalg.det)(_default_unary_wrapper(unyt_linalg_det))
implements(np.diff)(_default_unary_wrapper(unyt_diff))
implements(np.ediff1d)(_default_unary_wrapper(unyt_ediff1d))
implements(np.ptp)(_default_unary_wrapper(unyt_ptp))
# implements(np.cumprod)(...) Omitted because unyt just raises if called.
# implements(np.cumulative_prod)(...) Omitted because unyt just raises if called.
implements(np.pad)(_default_unary_wrapper(unyt_pad))
implements(np.choose)(_default_binary_wrapper(unyt_choose))
implements(np.insert)(_default_unary_wrapper(unyt_insert))


@implements(np.linalg.lstsq)
def linalg_lstsq(a, b, rcond=None):
    helper_result = _prepare_array_func_args(a, b, rcond=rcond)
    ress = unyt_linalg_lstsq(*helper_result["args"], **helper_result["kwargs"])
    return (
        _return_helper(ress[0], helper_result),
        _return_helper(ress[1], helper_result),
        ress[2],
        _return_helper(ress[3], helper_result),
    )


implements(np.linalg.solve)(_default_binary_wrapper(unyt_linalg_solve))
implements(np.linalg.tensorsolve)(_default_binary_wrapper(unyt_linalg_tensorsolve))


@implements(np.linalg.eig)
def linalg_eig(a):
    helper_result = _prepare_array_func_args(a)
    ress = unyt_linalg_eig(*helper_result["args"], **helper_result["kwargs"])
    return (_return_helper(ress[0], helper_result), ress[1])


@implements(np.linalg.eigh)
def linalg_eigh(a, UPLO="L"):
    helper_result = _prepare_array_func_args(a, UPLO=UPLO)
    ress = unyt_linalg_eigh(*helper_result["args"], **helper_result["kwargs"])
    return (_return_helper(ress[0], helper_result), ress[1])


implements(np.linalg.eigvals)(_default_unary_wrapper(unyt_linalg_eigvals))
implements(np.linalg.eigvalsh)(_default_unary_wrapper(unyt_linalg_eigvalsh))


@implements(np.savetxt)
def savetxt(
    fname,
    X,
    fmt="%.18e",
    delimiter=" ",
    newline="\n",
    header="",
    footer="",
    comments="# ",
    encoding=None,
):
    warnings.warn(
        "numpy.savetxt does not preserve units or extra_attr information, "
        "and will only save the raw numerical data from the subclass_uarray object.\n"
        "If this is the intended behaviour, call `numpy.savetxt(file, arr.d)` "
        "to silence this warning.\n",
        stacklevel=4,
    )
    helper_result = _prepare_array_func_args(
        fname,
        X,
        fmt=fmt,
        delimiter=delimiter,
        newline=newline,
        header=header,
        footer=footer,
        comments=comments,
        encoding=encoding,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            category=UserWarning,
            message="numpy.savetxt does not preserve units",
        )
        unyt_savetxt(*helper_result["args"], **helper_result["kwargs"])
    return


@implements(np.apply_over_axes)
def apply_over_axes(func, a, axes):
    res = func(a, axes[0])
    if len(axes) > 1:
        # this function is recursive by nature,
        # here we intentionally do not call the base _implementation
        return np.apply_over_axes(func, res, axes[1:])
    else:
        return res


implements(np.fill_diagonal)(_default_binary_wrapper(unyt_fill_diagonal))
implements(np.isin)(_default_comparison_wrapper(unyt_isin))
implements(np.place)(_default_unary_wrapper(unyt_place))
implements(np.put)(_default_unary_wrapper(unyt_put))
implements(np.put_along_axis)(_default_unary_wrapper(unyt_put_along_axis))
implements(np.putmask)(_default_unary_wrapper(unyt_putmask))
implements(np.searchsorted)(_default_binary_wrapper(unyt_searchsorted))


@implements(np.select)
def select(condlist, choicelist, default=0):
    helper_result = _prepare_array_func_args(condlist, choicelist, default=default)
    helper_result_choicelist = _prepare_array_func_args(*choicelist)
    res = unyt_select(
        helper_result["args"][0],
        helper_result_choicelist["args"],
        **helper_result["kwargs"],
    )
    return _return_helper(res, helper_result)


implements(np.setdiff1d)(_default_binary_wrapper(unyt_setdiff1d))
implements(np.sinc)(_default_unary_wrapper(unyt_sinc))
implements(np.clip)(_default_unary_wrapper(unyt_clip))


@implements(np.where)
def where(condition, *args):
    helper_result = _prepare_array_func_args(condition, *args)
    if len(args) == 0:  # just condition
        res = unyt_where(*helper_result["args"], **helper_result["kwargs"])
    elif len(args) < 2:
        # error message borrowed from numpy 1.24.1
        raise ValueError("either both or neither of x and y should be given")
    res = unyt_where(*helper_result["args"], **helper_result["kwargs"])
    return _return_helper(res, helper_result)


implements(np.triu)(_default_unary_wrapper(unyt_triu))
implements(np.tril)(_default_unary_wrapper(unyt_tril))


@implements(np.einsum)
def einsum(
    subscripts,
    *operands,
    out=None,
    dtype=None,
    order="K",
    casting="safe",
    optimize=False,
):
    helper_result = _prepare_array_func_args(
        subscripts,
        operands,
        out=out,
        dtype=dtype,
        order=order,
        casting=casting,
        optimize=optimize,
    )
    helper_result_operands = _prepare_array_func_args(*operands)
    res = unyt_einsum(
        helper_result["args"][0],
        *helper_result_operands["args"],
        **helper_result["kwargs"],
    )
    return _return_helper(res, helper_result_operands, out=out)


implements(np.convolve)(_default_binary_wrapper(unyt_convolve))
implements(np.correlate)(_default_binary_wrapper(unyt_correlate))
implements(np.tensordot)(_default_binary_wrapper(unyt_tensordot))
implements(np.unwrap)(_default_unary_wrapper(unyt_unwrap))
implements(np.interp)(_default_unary_wrapper(unyt_interp))


@implements(np.array_repr)
def array_repr(arr, max_line_width=None, precision=None, suppress_small=None):
    helper_result = _prepare_array_func_args(
        arr,
        max_line_width=max_line_width,
        precision=precision,
        suppress_small=suppress_small,
    )
    rep = unyt_array_repr(*helper_result["args"], **helper_result["kwargs"])[:-1]
    if hasattr(arr, "extra_attr"):
        rep += f", extra_attr='{arr.extra_attr}'"
    rep += ")"
    return rep


if NUMPY_VERSION < Version("2.0.0dev0"):
    implements(np.asfarray)(_default_unary_wrapper(unyt_asfarray))  # NOQA: NPY201
if NUMPY_VERSION >= Version("2.0.0dev0"):
    implements(np.linalg.outer)(_default_binary_wrapper(unyt_linalg_outer))
implements(_trapezoid_func)(_default_binary_wrapper(unyt_trapezoid))
implements(np.isin)(_default_comparison_wrapper(unyt_isin))
implements(np.take)(_default_unary_wrapper(unyt_take))

# Now we wrap functions that unyt does not handle explicitly:
implements(np.average)(_propagate_extra_attr_to_result(np.average._implementation))
implements(np.max)(_propagate_extra_attr_to_result(np.max._implementation))
implements(np.min)(_propagate_extra_attr_to_result(np.min._implementation))
implements(np.mean)(_propagate_extra_attr_to_result(np.mean._implementation))
implements(np.median)(_propagate_extra_attr_to_result(np.median._implementation))
implements(np.sort)(_propagate_extra_attr_to_result(np.sort._implementation))
implements(np.sum)(_propagate_extra_attr_to_result(np.sum._implementation))
implements(np.partition)(_propagate_extra_attr_to_result(np.partition._implementation))
if NUMPY_VERSION >= Version("2.0.0dev0"):
    implements(np.linalg.cross)(
        _default_binary_wrapper(np.linalg.cross._implementation)
    )


@implements(np.meshgrid)
def meshgrid(*xi, **kwargs):
    # meshgrid is a unique case: arguments never interact with each other, so we don't
    # want to use our _prepare_array_func_args helper (that will try to enforce
    # consistency).
    # However we can't just use _propagate_extra_attr_to_result because we
    # need to iterate over arguments.
    res = np.meshgrid._implementation(*xi, **kwargs)
    ret_type = tuple if NUMPY_VERSION >= Version("2.0.0dev0") else list
    return ret_type(_copy_extra_attr_if_present(x, r) for (x, r) in zip(xi, res))


class subclass_uarray(unyt_array):
    """
    Minimalist subclass of unyt_array.
    """

    def __new__(
        cls,
        input_array: Iterable,
        units: str | unyt.unit_object.Unit = None,
        *,
        registry: unyt.unit_registry.UnitRegistry = None,
        dtype: np.dtype | str = None,
        bypass_validation: bool = False,
        name: str = None,
        extra_attr: bool = None,
    ) -> "subclass_uarray":
        """
        Closely inspired by the :meth:`unyt.array.unyt_array.__new__` constructor.
        """

        if bypass_validation is True:
            obj = super().__new__(
                cls,
                input_array,
                units=units,
                registry=registry,
                dtype=dtype,
                bypass_validation=bypass_validation,
                name=name,
            )

            # dtype, units, registry & name are handled by unyt
            obj.extra_attr = extra_attr

            return obj

        if isinstance(input_array, subclass_uarray):
            if input_array.extra_attr != extra_attr:
                raise ValueError
            obj = input_array.view(cls)

            return obj

        elif isinstance(input_array, np.ndarray) and input_array.dtype != object:
            # Guard np.ndarray so it doesn't get caught by _iterable in next
            # case. ndarray with object dtype goes to next case to properly
            # handle e.g. ndarrays containing subclass_uquantity's
            pass

        elif _iterable(input_array) and input_array:
            # check compatibility:
            helper_result = _prepare_array_func_args(*input_array)
            input_array = helper_result["args"]

            extra_attr = (
                helper_result["extra_attr"] if extra_attr is None else extra_attr
            )

        obj = super().__new__(
            cls,
            input_array,
            units=units,
            registry=registry,
            dtype=dtype,
            bypass_validation=bypass_validation,
            name=name,
        )

        if isinstance(obj, unyt_array) and not isinstance(obj, cls):
            obj = obj.view(cls)

        # attach our attribute:
        obj.extra_attr = extra_attr

        return obj

    def __array_finalize__(self, obj: "subclass_uarray") -> None:
        super().__array_finalize__(obj)
        if obj is None:
            return
        self.extra_attr = getattr(obj, "extra_attr", None)

    def __str__(self) -> str:
        return super().__str__() + f" {self.extra_attr}"

    def __repr__(self) -> str:
        return super().__repr__() + f" {self.extra_attr}"

    def __reduce__(self) -> tuple:
        """
        Pickle reduction method.

        Here we add an extra element at the start of the state tuple to store
        the extra_attr.
        """
        np_ret = super().__reduce__()
        obj_state = np_ret[2]
        sub_state = (((self.extra_attr,),) + obj_state[:],)
        new_ret = np_ret[:2] + sub_state + np_ret[3:]
        return new_ret

    def __setstate__(self, state: tuple) -> None:
        """
        Pickle setstate method.

        Here we extract the extra info we added to the object
        state and pass the rest to :meth:`unyt.array.unyt_array.__setstate__`.
        """
        super().__setstate__(state[1:])
        (self.extra_attr,) = state[0]

    # Wrap functions that return copies of subclass_uarrays so that our
    # attribute gets passed through:
    astype = _propagate_extra_attr_to_result(unyt_array.astype)
    in_units = _propagate_extra_attr_to_result(unyt_array.in_units)
    byteswap = _propagate_extra_attr_to_result(unyt_array.byteswap)
    compress = _propagate_extra_attr_to_result(unyt_array.compress)
    diagonal = _propagate_extra_attr_to_result(unyt_array.diagonal)
    flatten = _propagate_extra_attr_to_result(unyt_array.flatten)
    ravel = _propagate_extra_attr_to_result(unyt_array.ravel)
    repeat = _propagate_extra_attr_to_result(unyt_array.repeat)
    swapaxes = _propagate_extra_attr_to_result(unyt_array.swapaxes)
    transpose = _propagate_extra_attr_to_result(unyt_array.transpose)
    view = _propagate_extra_attr_to_result(unyt_array.view)
    __copy__ = _propagate_extra_attr_to_result(unyt_array.__copy__)
    __deepcopy__ = _propagate_extra_attr_to_result(unyt_array.__deepcopy__)
    in_cgs = _propagate_extra_attr_to_result(unyt_array.in_cgs)
    take = _propagate_extra_attr_to_result(
        _ensure_result_is_subclass_uarray_or_uquantity(unyt_array.take)
    )
    reshape = _propagate_extra_attr_to_result(
        _ensure_result_is_subclass_uarray_or_uquantity(unyt_array.reshape)
    )
    __getitem__ = _propagate_extra_attr_to_result(
        _ensure_result_is_subclass_uarray_or_uquantity(unyt_array.__getitem__)
    )
    dot = _default_binary_wrapper(unyt_array.dot)

    # Also wrap some array "properties":
    T = property(_propagate_extra_attr_to_result(unyt_array.transpose))
    ua = property(_propagate_extra_attr_to_result(np.ones_like))
    unit_array = property(_propagate_extra_attr_to_result(np.ones_like))

    @classmethod
    def __unyt_ufunc_prepare__(cls, ufunc: np.ufunc, method: str, *inputs, **kwargs):
        helper_result = _prepare_array_func_args(*inputs, **kwargs)
        return ufunc, method, helper_result["args"], helper_result["kwargs"]

    @classmethod
    def __unyt_ufunc_finalize__(
        cls, result, ufunc: np.ufunc, method: str, *inputs, **kwargs
    ):
        helper_result = _prepare_array_func_args(*inputs, **kwargs)
        # if we get a tuple we have multiple return values to deal with
        if isinstance(result, tuple):
            r = tuple(
                (
                    r.view(subclass_uquantity)
                    if r.shape == ()
                    else (
                        r.view(subclass_uarray)
                        if isinstance(r, unyt_array)
                        and not isinstance(r, subclass_uarray)
                        else r
                    )
                )
                for r in result
            )
            for r in result:
                if isinstance(r, subclass_uarray):  # also recognizes subclass_uquantity
                    r.extra_attr = helper_result["extra_attr"]
        elif isinstance(result, unyt_array):  # also recognizes subclass_uquantity
            if not isinstance(result, subclass_uarray):
                result = (
                    result.view(subclass_uquantity)
                    if result.shape == ()
                    else result.view(subclass_uarray)
                )
            result.extra_attr = helper_result["extra_attr"]
        if "out" in kwargs:
            out = kwargs.pop("out")
            if ufunc not in multiple_output_operators:
                out = out[0]
                if isinstance(out, unyt_array) and not isinstance(out, subclass_uarray):
                    out = (
                        out.view(subclass_uquantity)
                        if out.shape == ()
                        else out.view(subclass_uarray)
                    )
                if isinstance(
                    out, subclass_uarray
                ):  # also recognizes subclass_uquantity
                    out.extra_attr = helper_result["extra_attr"]
            else:
                out = tuple(
                    (
                        (
                            o.view(subclass_uquantity)
                            if o.shape == ()
                            else o.view(subclass_uarray)
                        )
                        if isinstance(o, unyt_array)
                        and not isinstance(o, subclass_uarray)
                        else o
                    )
                    for o in out
                )
                for o in out:
                    if isinstance(
                        o, subclass_uarray
                    ):  # also recognizes subclass_uquantity
                        o.extra_attr = helper_result["extra_attr"]
        return result

    def __array_ufunc__(
        self, ufunc: np.ufunc, method: str, *inputs, **kwargs
    ) -> object:
        """
        Handles :mod:`numpy` ufunc calls on ``subclass_uarray`` input.

        ``numpy`` facilitates wrapping array classes by handing off to this
        function when a function of :class:`numpy.ufunc` type is called with
        arguments from an inheriting array class. Since we inherit from
        ``unyt_array`, we let it handle what to do with the units and
        take care of processing the extra_attr via our helper functions.
        """
        helper_result = _prepare_array_func_args(*inputs, **kwargs)

        result = _ensure_result_is_subclass_uarray_or_uquantity(
            super().__array_ufunc__
        )(ufunc, method, *inputs, **kwargs)
        # if we get a tuple we have multiple return values to deal with
        if isinstance(result, tuple):
            for r in result:
                if isinstance(r, subclass_uarray):  # also recognizes subclass_uquantity
                    r.extra_attr = helper_result["extra_attr"]
        elif isinstance(result, subclass_uarray):  # also recognizes subclass_uquantity
            result.extra_attr = helper_result["extra_attr"]
        if "out" in kwargs:
            out = kwargs.pop("out")
            if ufunc not in multiple_output_operators:
                out = out[0]
                if isinstance(
                    out, subclass_uarray
                ):  # also recognizes subclass_uquantity
                    out.extra_attr = helper_result["extra_attr"]
            else:
                for o in out:
                    if isinstance(
                        o, subclass_uarray
                    ):  # also recognizes subclass_uquantity
                        o.extra_attr = helper_result["extra_attr"]

        return result

    def __array_function__(
        self, func: Callable, types: Collection, args: tuple, kwargs: dict
    ):
        """
        Handles ``numpy`` function calls on ``subclass_uarray`` input.

        ``numpy`` facilitates wrapping array classes by handing off to this
        function when a numpy-defined function is called with arguments
        from an inheriting array class. Since we inherit from ``unyt_array``,
        we let it handle what to do with the units and take care of processing
        the extra_attr via our helper functions.
        """
        # Follow NEP 18 guidelines
        # https://numpy.org/neps/nep-0018-array-function-protocol.html
        from unyt._array_functions import (
            _HANDLED_FUNCTIONS as _UNYT_HANDLED_FUNCTIONS,
            _UNSUPPORTED_FUNCTIONS as _UNYT_UNSUPPORTED_FUNCTIONS,
        )

        # Let's claim to support everything supported by unyt.
        # If we can't do this in future, follow their pattern of
        # defining out own _UNSUPPORTED_FUNCTIONS in a _array_functions.py file
        _UNSUPPORTED_FUNCTIONS = _UNYT_UNSUPPORTED_FUNCTIONS

        if func in _UNSUPPORTED_FUNCTIONS:
            # following NEP 18, return NotImplemented as a sentinel value
            # which will lead to raising a TypeError, while
            # leaving other arguments a chance to take the lead
            return NotImplemented

        if not all(issubclass(t, subclass_uarray) or t is np.ndarray for t in types):
            # Note: this allows subclasses that don't override
            # __array_function__ to handle subclass_uarray objects
            return NotImplemented

        if func in _HANDLED_FUNCTIONS:
            function_to_invoke = _HANDLED_FUNCTIONS[func]
        elif func in _UNYT_HANDLED_FUNCTIONS:
            function_to_invoke = _UNYT_HANDLED_FUNCTIONS[func]
        else:
            # default to numpy's private implementation
            function_to_invoke = func._implementation
        return function_to_invoke(*args, **kwargs)

    def __mul__(
        self, b: int | float | np.ndarray | unyt.unit_object.Unit
    ) -> "subclass_uarray":
        """
        Multiply this ``subclass_uarray``.

        We delegate most cases to ``unyt_array``, but we need to handle
        the case where the second argument is a ``Unit``.
        """
        if getattr(b, "is_Unit", False):
            return _copy_extra_attr_if_present(
                self,
                _ensure_result_is_subclass_uarray_or_uquantity(b.__mul__)(
                    self.view(unyt_quantity)
                    if self.shape == ()
                    else self.view(unyt_array)
                ),
            )
        else:
            return super().__mul__(b)

    def __rmul__(
        self, b: int | float | np.ndarray | unyt.unit_object.Unit
    ) -> "subclass_uarray":
        """
        Multiply this ``subclass_uarray`` (as the right argument).

        We delegate most cases to ``unyt_array``, but we need to handle
        the case where the second argument is a ``Unit``.
        """
        if getattr(b, "is_Unit", False):
            return self.__mul__(b)
        else:
            return super().__rmul__(b)


class subclass_uquantity(subclass_uarray, unyt_quantity):
    """
    Minimalist subclass of unyt_quantity.
    """

    def __new__(
        cls,
        input_scalar: numeric_type,
        units: str | unyt.unit_object.Unit | None = None,
        *,
        registry: unyt.unit_registry.UnitRegistry | None = None,
        dtype: np.dtype | str | None = None,
        bypass_validation: bool = False,
        name: str | None = None,
        extra_attr: bool | None = None,
    ) -> "subclass_uquantity":
        """
        Closely inspired by the ``unyt_quantity.__new__`` constructor.
        """
        if bypass_validation is True:
            result = super().__new__(
                cls,
                np.asarray(input_scalar),
                units=units,
                registry=registry,
                dtype=dtype,
                bypass_validation=bypass_validation,
                name=name,
                extra_attr=extra_attr,
            )

        if not isinstance(input_scalar, (numeric_type, np.number, np.ndarray)):
            raise RuntimeError("subclass_uquantity values must be numeric")

        # Use values from kwargs, if None use values from input_scalar
        units = getattr(input_scalar, "units", None) if units is None else units
        name = getattr(input_scalar, "name", None) if name is None else name
        if isinstance(input_scalar, subclass_uquantity):
            if input_scalar.extra_attr != extra_attr:
                raise ValueError
        extra_attr = (
            getattr(input_scalar, "extra_attr", None)
            if extra_attr is None
            else extra_attr
        )
        result = super().__new__(
            cls,
            np.asarray(input_scalar),
            units=units,
            registry=registry,
            dtype=dtype,
            bypass_validation=bypass_validation,
            name=name,
            extra_attr=extra_attr,
        )
        if result.size > 1:
            raise RuntimeError("subclass_uquantity instances must be scalars")
        return result

    __round__ = _propagate_extra_attr_to_result(
        _ensure_result_is_subclass_uarray_or_uquantity(unyt_quantity.__round__)
    )
